import ctypes
import os
import logging as lg

import pyqtgraph as pg
from PyQt6.QtWidgets import QApplication, QMainWindow, QPlainTextEdit
from pyqtgraph.Qt import QtGui, QtWidgets
from pyqtgraph.dockarea import *
import numpy as np
import toml
import fringes as frng

from fringes_gui.setters import set_functionality

# import pyqtgraph.examples
# pyqtgraph.examples.run()


class FringesGUI(QApplication):
    """Simple graphical user interface for the 'fringes' package."""

    def __init__(self):
        super(FringesGUI, self).__init__([])

        myappid = "Fringes-GUI"  # arbitrary string
        try:
            version = toml.load("..\\pyproject.toml")["tool"]["poetry"]["version"]
            myappid += "_" + version
        except Exception:
            pass
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

        pg.setConfigOptions(imageAxisOrder="row-major", useNumba=True)  # useCupy

        self.fringes = frng.Fringes(X=1920, Y=1200)
        self.fringes.logger.setLevel("INFO")
        fname = os.path.join(os.path.dirname(__file__), "params.yaml")
        if os.path.isfile(fname):
            self.fringes.load(fname)
        self.initials = self.fringes.params
        self.key = ""
        self.visibility = "Expert"
        self.digits = 8  # todo: len(str(self.fringes._Pmax))  # 4 (digits) + 1 (point) + 3 (decimals) = 8 == current length of Pmax?
        self.sub = str.maketrans("1234567890", "₁₂₃₄₅₆₇₈₉₀")
        self.sup = str.maketrans("₁₂₃₄₅₆₇₈₉₀", "1234567890")

        # define window
        self.win = QMainWindow()
        self.area = DockArea()
        self.win.setCentralWidget(self.area)

        self.win.setGeometry(QtGui.QGuiApplication.primaryScreen().availableGeometry())  # move to primary screen
        # if Screen.count == 1 and self.desktop.availableGeometry(idx).width() / self.desktop.availableGeometry(idx).height() >= 21 / 9:
        #     self.win.resize(self.desktop.availableGeometry().width() // 2, self.desktop.availableGeometry().height())
        #     # self.win.move(0, 0)  # move to the left
        #     self.win.move(self.desktop.availableGeometry().width() // 2, 0)  # move to the right
        # else:
        #     self.win.showMaximized()  # self.win.showFullScreen()

        self.win.showMaximized()  # self.win.showFullScreen()

        self.win.setWindowIcon(QtGui.QIcon("spirals.png"))

        self.win.setWindowTitle(myappid)

        # set style
        # self.setStyleSheet(open("QTDark.stylesheet").read())

        # import qdarkstyle
        # self.setStyleSheet(qdarkstyle.load_stylesheet(qt_api='pyqt6'))

        # styles = QtWidgets.QStyleFactory.keys()  # system styles
        # self.setStyle('Fusion')

        # import qtmodern.styles
        # import qtmodern.windows
        # qtmodern.styles.dark(self)
        # self.win = qtmodern.windows.ModernWindow(self.win)  # todo: moving window with win+arrow keys doesn't work
        # pg.setConfigOptions(background=42/255)

        # Create docks, place them into the window one at a time.
        from pyqtgraph.dockarea.Dock import DockLabel

        def updateStylePatched(self):  # from https://gist.github.com/matmr/72487a03da95b99db6ae
            r = '3px'
            if self.dim:
                fg = '#fff'  # gray
                bg = "##6CC0A8"  # green brighter
                border = "#6CC0A8"  # green brighter
            else:
                fg = '#fff'  # white
                bg = "#169D7C"  # green
                border = "#169D7C"  # green

            if self.orientation == 'vertical':
                self.vStyle = """DockLabel {
                    background-color : %s;
                    color : %s;
                    border-top-right-radius: 0px;
                    border-top-left-radius: %s;
                    border-bottom-right-radius: 0px;
                    border-bottom-left-radius: %s;
                    border-width: 0px;
                    border-right: 2px solid %s;
                    padding-top: 3px;
                    padding-bottom: 3px;
                }""" % (bg, fg, r, r, border)
                self.setStyleSheet(self.vStyle)
            else:
                self.hStyle = """DockLabel {
                    background-color : %s;
                    color : %s;
                    border-top-right-radius: %s;
                    border-top-left-radius: %s;
                    border-bottom-right-radius: 0px;
                    border-bottom-left-radius: 0px;
                    border-width: 0px;
                    border-bottom: 2px solid %s;
                }""" % (bg, fg, r, r, border)
                self.setStyleSheet(self.hStyle)

        DockLabel.updateStyle = updateStylePatched

        self.dock_attributes = Dock("Attributes", size=(15, 99))
        self.dock_methods = Dock("Methods", size=(15, 1))
        self.dock_viewer = Dock("Viewer", size=(70, 100))
        self.dock_data = Dock("Data", size=(15, 30))
        self.dock_log = Dock("Log", size=(15, 70))

        self.area.addDock(self.dock_attributes, "left")
        self.area.addDock(self.dock_viewer, "right", self.dock_attributes)
        self.area.addDock(self.dock_methods, "bottom", self.dock_attributes)
        self.area.addDock(self.dock_data, "right", self.dock_viewer)
        self.area.addDock(self.dock_log, "bottom", self.dock_data)

        # Add widgets into each dock.

        # General settings
        self.immerse = False
        self.immerse_key = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+Shift+i"), self.win)

        self.tree = pg.parametertree.ParameterTree()
        self.dock_attributes.addWidget(self.tree)

        # self.reset_button = QtWidgets.QPushButton("Reset")  # todo
        # self.reset_button.setToolTip("Reset parameters to values on start up.")
        # self.dock_attributes.addWidget(self.reset_button)
        self.reset_key = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+R"), self.win)

        # self.undo_key = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+Z"), self.win)  # todo
        # self.redo_key = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+Y"), self.win)

        # Control
        self.encode_checkbox = QtWidgets.QCheckBox()
        self.encode_label = QtWidgets.QLabel("      Encode on parameter change")
        self.decode_checkbox = QtWidgets.QCheckBox()
        self.decode_label = QtWidgets.QLabel("      Decode on parameter change")
        self.reset_button = QtWidgets.QPushButton("Reset")
        self.reset_button.setEnabled(False)
        self.reset_button.setToolTip("Press 'Ctrl+R'.")
        self.default_key = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+Shift+D"), self.win)
        self.encode_button = QtWidgets.QPushButton("Encode")
        self.encode_button.setToolTip("Press 'E'.")
        self.encode_button.setStyleSheet("" if self.encodeOK else "QPushButton{color: red}")
        self.encode_key = QtGui.QShortcut(QtGui.QKeySequence("E"), self.win)
        self.decode_button = QtWidgets.QPushButton("Decode")
        self.decode_button.setEnabled(False)
        self.decode_button.setToolTip("Press 'D'.")
        self.decode_key = QtGui.QShortcut(QtGui.QKeySequence("D"), self.win)
        self.decode_key.setEnabled(False)
        self.remap_button = QtWidgets.QPushButton("Remap")
        self.remap_button.setEnabled(False)
        self.remap_button.setToolTip("Press 'R'.")
        self.remap_key = QtGui.QShortcut(QtGui.QKeySequence("R"), self.win)
        self.remap_key.setEnabled(False)
        self.curvature_button = QtWidgets.QPushButton("Curvature")
        self.curvature_button.setEnabled(False)
        self.curvature_button.setToolTip("Press 'C'.")
        self.curvature_key = QtGui.QShortcut(QtGui.QKeySequence("C"), self.win)
        self.curvature_key.setEnabled(False)
        self.height_button = QtWidgets.QPushButton("Height")
        self.height_button.setEnabled(False)
        self.height_button.setToolTip("Press 'H'.")
        self.height_key = QtGui.QShortcut(QtGui.QKeySequence("H"), self.win)
        self.height_key.setEnabled(False)
        self.dock_methods.addWidget(self.reset_button, 0, 0, 1, 2)
        # self.dock_methods.addWidget(self.encode_label, 1, 0, 1, 2)
        # self.dock_methods.addWidget(self.encode_checkbox, 1, 0, 1, 2)
        # self.dock_methods.addWidget(self.decode_label, 2, 0, 1, 2)
        # self.dock_methods.addWidget(self.decode_checkbox, 2, 0, 1, 2)
        self.dock_methods.addWidget(self.encode_button, 3, 0)
        self.dock_methods.addWidget(self.decode_button, 3, 1)
        self.dock_methods.addWidget(self.remap_button, 4, 0, 1, 2)
        self.dock_methods.addWidget(self.curvature_button, 5, 0)
        self.dock_methods.addWidget(self.height_button, 5, 1)

        # Viewer
        self.zoomback_key = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+0"), self.win)  # todo

        self.plot = pg.PlotItem()
        self.plot.setLabel(axis='left', text='y-axis')
        self.plot.setLabel(axis='bottom', text='x-axis')  # todo: set label 'T-axis'
        self.imv = pg.ImageView(view=self.plot)
        # self.imv.ui.histogram.hide()
        # self.imv.ui.roiBtn.hide()
        # self.imv.ui.menuBtn.hide()
        self.dock_viewer.addWidget(self.imv)

        self.info = QtWidgets.QLabel()
        self.dock_viewer.addWidget(self.info)

        # Data
        # self.data_tree = pg.DataTreeWidget()

        class TableView(QtWidgets.QTableWidget):
            def __init__(self, *args):
                QtWidgets.QTableWidget.__init__(self, *args)
                self.resizeRowsToContents()
                self.resizeColumnsToContents()
                # self.setSelectionBehavior(QtWidgets.QTableView.selectRow)
                # self.setSelectionMode(QtWidgets.QTableView.SingleSelection)
                self.setColumnCount(3)
                self.setRowCount(0)
                self.setHorizontalHeaderLabels(["Name", "Video-Shape", "Type"])

            def setData(self, data={}):
                self.setRowCount(len(data))
                for i, row in enumerate(sorted(data)):
                    for j, v in enumerate(row):
                        newitem = QtWidgets.QTableWidgetItem(v)
                        self.setItem(i, j, newitem)

                # self.resizeRowsToContents()
                self.resizeColumnsToContents()

        # self.display_selector = QtWidgets.QComboBox()
        # self.display_selector.setPlaceholderText("Nothing")
        # self.con = Container(self.display_selector)
        class Container:
            @property
            def info(self):
                info = [[str(k), str(v.shape), str(v.dtype)] for k, v in self.__dict__.items() if
                        isinstance(v, np.ndarray)]
                return info

        self.con = Container()
        # self.dock_data.addWidget(self.display_selector, 0, 0)

        self.data_table = TableView()
        self.dock_data.addWidget(self.data_table, 1, 0, 1, 2)
        self.load_button = QtWidgets.QPushButton("Load")
        self.load_button.setToolTip("Press 'Ctrl+L'.")
        self.load_key = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+L"), self.win)
        self.save_button = QtWidgets.QPushButton("Save")
        self.save_button.setToolTip("Press 'Ctrl+S'.")
        self.save_key = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+S"), self.win)
        self.set_button = QtWidgets.QPushButton("Set data (to be decoded)")
        self.set_button.setToolTip("Press 'Ctrl+Shift+S'.")
        self.set_button.setEnabled(False)
        self.set_key = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+Shift+S"), self.win)
        self.clear_button = QtWidgets.QPushButton("Clear all")
        self.clear_button.setEnabled(self.dataOK)
        self.clear_button.setToolTip("Press 'Ctrl+Shift+C'.")
        self.clear_key = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+Shift+C"), self.win)
        self.clear_key.setEnabled(self.dataOK)
        self.dock_data.addWidget(self.load_button, 2, 0)
        self.dock_data.addWidget(self.save_button, 2, 1)
        self.dock_data.addWidget(self.clear_button, 3, 0)
        self.dock_data.addWidget(self.set_button, 3, 1)

        # self.model = QtWidgets.QFileSystemModel()
        # self.model.setRootPath(self.data.dest)
        # self.view = QtWidgets.QTreeView()
        # self.view.setModel(self.model)
        # self.view.setCurrentIndex(self.model.index(self.data.dest))
        # self.view.setExpanded(self.model.index(self.data.dest), True)
        # self.dock_data.addWidget(self.view, row=3, col=0)

        # Log
        class QPlainTextEditLogger(lg.Handler):
            def emit(self, record):
                self.widget.appendPlainText(self.format(record))

        self.log_widget = QPlainTextEdit()
        self.dock_log.addWidget(self.log_widget)
        handler = QPlainTextEditLogger()
        handler.setFormatter(self.fringes.logger.handlers[0].formatter)
        handler.widget = self.log_widget
        self.fringes.logger.addHandler(handler)

        # todo: reset button or simply restart? save current config on shutdown

        set_functionality(self)

        self.show()
        # self.fringes.logger.info(f"Started {myappid}.")

    def show(self):
        """Display the Application."""
        pg.exec()
        # if (sys.flags.interactive != 1) or not hasattr(QtCore, "PYQT_VERSION"):
        #     QtWidgets.QApplication.instance().exec_()

    @property
    def SDMOK(self):
        a = self.fringes.D == 2
        b = self.fringes.grid in self.fringes._grids[:2]
        c = not self.fringes.FDM
        return a and b and c

    @property
    def WDMOK(self):
        a = self.fringes._ismono
        b = np.all(self.fringes.N == 3)
        e = not self.fringes.FDM
        return a and b and e

    @property
    def FDMOK(self):
        a = self.fringes.D > 1 or self.fringes.K > 1
        b = not self.fringes.SDM
        c = not self.fringes.WDM
        return a and b and c

    @property
    def muxtxt(self):
        return "Single Shot" if self.fringes.T == 1 else "Crossed Fringes" if self.fringes.D == 2 and (self.fringes.SDM or self.fringes.FDM) else ""

    @property
    def encodeOK(self):
        """True if unambiguous measurement range >= length."""
        return np.all(self.fringes.UMR >= self.fringes.L)

    @property
    def dataOK(self):
        """Return True if at least one attribute of Container class is an ndarray object."""
        return any(isinstance(obj, np.ndarray) for obj in self.con.__dict__.values())

    @property
    def set_dataOK(self):
        """True if data to be decoded can be set."""
        flist = [v.shape[0] for k, v in self.con.__dict__.items() if isinstance(v, np.ndarray)]
        return any(frames == self.fringes.T for frames in flist)

    @property
    def decodeOK(self):
        """Return true if data present can be decoded."""
        I = (
            getattr(self.con, self.key) if hasattr(self.con, self.key)
            # else self.con.raw if hasattr(self.con, "raw")
            else self.con.fringes if hasattr(self.con, "fringes")
            else None
        )
        return I is not None and hasattr(I, "ndim") and frng.vshape(I).shape[0] == self.fringes.T

    @property
    def remapOK(self):
        """Returns True if modulation and registration is available."""
        return hasattr(self.con, "registration")  # hasattr(self.con, "modulation")

    @property
    def curvatureOK(self):
        return hasattr(self.con, "registration") and self.con.registration.shape[1] >= 2 and self.con.registration.shape[2] >= 2

    @property
    def heightOK(self):
        return hasattr(self.con, "curvature")

    @property
    def adv_vis(self):  # todo: remove
        """Returns True if the user has at least Expert level access i.e. advanced visibility."""
        return self.visibility in ["Expert", "Guru", "Experimental"]



import numpy as np
import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets

from .params import set_params
from .logic import set_logic


def set_functionality(gui):
    """Give the list of parameters defined in Gui their functionality, i.e. set listeners for when they are changed."""

    set_params(gui)

    gui.params = pg.parametertree.Parameter.create(name="Settings", type="group", children=gui.params)
    gui.tree.setParameters(gui.params, showTop=False)
    gui.params_initial = gui.params.saveState()

    set_logic(gui)

    vis = gui.visibility
    if gui.visibility != "Guru":
        gui.params.param("vis").setValue(
            "Guru")  # switching back from Guru mode prevents loading params which don't fit user options
    gui.params.param("vis").setValue(vis)

    def set_param(param, change):
        # param: The immediate child whose tree state has changed.
        #        Note that the change may have originated from a grandchild.
        # changes: List of tuples describing all changes that have been made in this event: (param, changeDescr, data)

        if change[0][1] != "value":
            return

        key = change[0][0].opts["name"]
        val = change[0][2]

        if key not in ["mode", "verbose"] or \
                key == "axis" and gui.fringes.D == 2:
            if hasattr(gui.con, "fringes"):
                delattr(gui.con, "fringes")

                if hasattr(gui.con, "registration"):
                    gui.view(gui.con.registration)
                elif hasattr(gui.con, gui.key):
                    gui.view(getattr(gui.con, gui.key))
                else:
                    gui.view(None)

            if gui.encode_checkbox.isChecked():
                gui.encode()

        if key not in ["alpha", "dtype", "bias", "amplitude", "visibility", "gamma"] or \
                key == "axis" and gui.fringes.D == 2:
            if hasattr(gui.con, "brightness"):
                delattr(gui.con, "brightness")

            if hasattr(gui.con, "modulation"):
                delattr(gui.con, "modulation")

            if hasattr(gui.con, "registration"):
                delattr(gui.con, "registration")

            if hasattr(gui.con, "phase"):
                delattr(gui.con, "phase")

            if hasattr(gui.con, "orders"):
                delattr(gui.con, "orders")

            if hasattr(gui.con, "residuals"):
                delattr(gui.con, "residuals")

            if hasattr(gui.con, gui.key):
                gui.view(getattr(gui.con, gui.key))
            elif hasattr(gui.con, "fringes"):
                gui.view(gui.con.fringes)
            else:
                gui.view(None)

            if gui.decode_checkbox.isChecked() and gui.decodeOK:
                gui.decode()

            gui.data_table.setData(gui.con.info)
            QtWidgets.QApplication.processEvents()  # refresh event queue

        if key == "vis":
            set_user(val)
            gui.fringes.logger.debug(f"Set visibility to {gui.visibility}.")
            return
        elif key == "log":
            gui.fringes.logger.setLevel(val)
            gui.fringes.logger.debug(f"Set logging level to '{val}'.")
            return

        if key[0] in "Nlvfh" and (key[-1] == "ᵢ" or key[-1].translate(gui.sup).isdigit()) and 2 <= len(key) <= 4:
            key, id = key[0], key[1:]
            idx = id.translate(gui.sup).split(",")
            d = int(idx[0]) if len(idx) == 2 else 0
            k = int(idx[-1]) if idx[-1] != "ᵢ" else 0  # if idx[-1] != "" else 0

            # workaround for problem of changing an element of a property object
            val_old = getattr(gui.fringes, "_" + key).copy()
            if key == "N":
                if val == 1:
                    val_old[:, :] = 1  # enable SSB
                elif gui.fringes.FDM:
                    val_old[:, :] = val
                elif gui.visibility in ["Guru", "Experimental"] or gui.fringes.FDM:
                    val_old[d, k] = val
                else:
                    val_old[:, k] = val
            elif key == "h":
                if val.red() == val.green() == val.blue() == 0:    # handle forbidden value 'black': set back to previous value
                    gui.params.param("col", "H", "h" + id).setValue(gui.fringes.h[k])
                    return
                else:
                    val_old[k] = [val.red(), val.green(), val.blue()]
            elif gui.visibility in ["Guru", "Experimental"] or gui.fringes.FDM:
                val_old[d, k] = val
            else:
                val_old[:, k] = val

            val = val_old
        elif key == "o":
            val *= np.pi

        setattr(gui.fringes, key, val)

        update_parameter_tree()

    def set_user(val):
        user_old = gui.visibility  # used if setting user to experimental
        gui.visibility = val

        with gui.params.treeChangeBlocker():
            if gui.visibility == "Beginner":
                gui.params.param("vid", "T").setLimits((3, gui.fringes._Tmax))
                gui.params.param("vid", "C").hide()
                gui.params.param("vid", "alpha").hide()
                gui.params.param("sys", "grid").setValue(gui.fringes.defaults["grid"])
                gui.params.param("sys", "grid").hide()
                gui.params.param("sys", "angle").setValue(gui.fringes.defaults["angle"])
                gui.params.param("sys", "angle").hide()
                gui.fringes.N = np.maximum(gui.fringes._N, 3)
                set_frequencies()
                gui.params.param("set", "f").hide()
                gui.params.param("set", "reverse").setValue(gui.fringes.defaults["reverse"])
                gui.params.param("set", "reverse").hide()
                gui.params.param("set", "o").setValue(gui.fringes.defaults["o"] / np.pi)
                gui.params.param("set", "o").hide()
                gui.params.param("set", "lmin").setValue(gui.fringes.defaults["lmin"])
                gui.params.param("set", "lmin").hide()
                gui.params.param("set", "vmax").hide()
                gui.params.param("val").hide()
                gui.params.param("val", "dtype").setValue(gui.fringes.defaults["dtype"])
                gui.params.param("val", "A").setValue(gui.fringes.defaults["A"])
                gui.params.param("val", "B").setValue(gui.fringes.defaults["B"])
                gui.params.param("val", "V").setValue(gui.fringes.defaults["V"])
                gui.params.param("val", "gamma").setValue(gui.fringes.defaults["gamma"])
                gui.params.param("mux").hide()
                gui.params.param("mux", "SDM").setValue(gui.fringes.defaults["SDM"])
                gui.params.param("mux", "WDM").setValue(gui.fringes.defaults["WDM"])
                gui.params.param("mux", "FDM", "static").setValue(gui.fringes.defaults["static"])
                gui.params.param("mux", "TDM").setValue(gui.fringes.TDM)
                gui.params.param("col").hide()
                gui.fringes.H = 1
                gui.params.param("uwr", "mode").setValue(gui.fringes.defaults["mode"])
                gui.params.param("uwr", "mode").hide()
                gui.params.param("uwr", "Vmin").setValue(gui.fringes.defaults["Vmin"])
                gui.params.param("uwr", "Vmin").hide()
                gui.params.param("uwr", "verbose").setValue(gui.fringes.defaults["verbose"])
                gui.params.param("uwr", "verbose").hide()
                gui.params.param("quali").hide()
                gui.params.param("quali", "dark").setValue(gui.fringes.defaults["dark"])
            elif gui.visibility == "Expert":
                gui.params.param("vid", "T").setLimits((3, gui.fringes._Tmax))
                gui.params.param("vid", "alpha").hide()
                gui.params.param("sys", "grid").setValue(gui.fringes.defaults["grid"])
                gui.params.param("sys", "grid").hide()
                gui.params.param("sys", "angle").setValue(gui.fringes.defaults["angle"])
                gui.params.param("sys", "angle").hide()
                gui.fringes.N = np.maximum(gui.fringes._N, 2)
                set_frequencies()
                gui.params.param("set", "f").hide()
                gui.params.param("set", "reverse").show()
                gui.params.param("set", "o").show()
                gui.params.param("set", "lmin").show()
                gui.params.param("set", "vmax").hide()
                gui.params.param("val").show()
                gui.params.param("val", "dtype").setValue(gui.fringes.defaults["dtype"])
                gui.params.param("val", "dtype").hide()
                gui.params.param("val", "Imax").setValue(gui.fringes.Imax)
                gui.params.param("val", "Imax").hide()
                gui.params.param("val", "A").show()
                gui.params.param("val", "B").show()
                gui.params.param("val", "V").show()
                gui.params.param("val", "gamma").show()
                gui.params.param("mux").hide()
                gui.params.param("mux", "SDM").setValue(gui.fringes.defaults["SDM"])
                gui.params.param("mux", "WDM").setValue(gui.fringes.defaults["WDM"])
                gui.params.param("mux", "FDM", "static").setValue(gui.fringes.defaults["static"])
                gui.params.param("mux", "TDM").setValue(gui.fringes.TDM)
                gui.params.param("col").hide()
                gui.fringes.M = 1
                gui.params.param("uwr", "mode").setValue(gui.fringes.defaults["mode"])
                gui.params.param("uwr", "mode").hide()
                gui.params.param("uwr", "Vmin").setValue(gui.fringes.defaults["Vmin"])
                gui.params.param("uwr", "Vmin").hide()
                gui.params.param("uwr", "verbose").show()
                gui.params.param("quali").show()
                gui.params.param("quali", "dark").hide()  # Experimental
                gui.params.param("quali", "shot").hide()  # Experimental
            elif gui.visibility == "Guru":
                gui.params.param("vid", "T").setLimits((1, gui.fringes._Tmax))
                gui.params.param("sys", "grid").setValue(gui.fringes.defaults["grid"])
                gui.params.param("sys", "grid").hide()  # todo: experimantal -> guru
                gui.params.param("sys", "angle").setValue(gui.fringes.defaults["angle"])
                gui.params.param("sys", "angle").hide()  # todo: experimantal -> guru
                gui.params.param("vid", "alpha").show()
                gui.params.param("set", "K").setLimits((1, (gui.fringes._Nmax - 1) / 2 / gui.fringes.D if gui.fringes.FDM else gui.fringes._Kmax))
                gui.params.param("set", "f").show()
                # gui.params.param("set", "vmax").show()
                gui.params.param("val", "dtype").show()
                gui.params.param("val", "Imax").show()
                gui.params.param("col", "H").setLimits((1, gui.fringes._Hmax))
                gui.params.param("col", "H").show()
                gui.params.param("mux").show()
                gui.params.param("col").show()
                gui.params.param("uwr", "mode").setValue(gui.fringes.defaults["mode"])
                gui.params.param("uwr", "mode").hide()
                gui.params.param("uwr", "Vmin").setValue(gui.fringes.defaults["Vmin"])
                gui.params.param("uwr", "Vmin").hide()
                gui.params.param("quali").show()
                gui.params.param("quali", "dark").hide()  # Experimental
                gui.params.param("quali", "shot").hide()  # Experimental
            elif gui.visibility == "Experimental":
                if user_old not in ["Guru", "Experimental"]:
                    gui.params.param("vis").setValue("Guru")  # get the settings from Guru mode first
                    gui.params.param("vis").setValue("Experimental")
                gui.params.param("sys", "grid").show()
                gui.params.param("sys", "angle").show()
                gui.params.param("uwr", "mode").show()
                gui.params.param("uwr", "Vmin").show()
                gui.params.param("quali", "dark").show()
                gui.params.param("quali", "shot").show()

            update_parameter_tree()

    def update_parameter_tree():
        is2D = max(gui.fringes.N.ndim, gui.fringes.l.ndim, gui.fringes.v.ndim, gui.fringes.f.ndim) == 2 or \
               gui.fringes.FDM or gui.visibility in ["Guru", "Experimental"]

        children = gui.fringes.D * gui.fringes.K if is2D else gui.fringes.K

        change_indices = len(gui.params.param("set", "v").children()) != children or \
                         gui.fringes.D != gui.params.param("sys", "D").value() or \
                         gui.fringes.K != gui.params.param("set", "K").value() or \
                         gui.fringes.FDM != gui.params.param("mux", "FDM").value() or \
                         gui.params.param("vis").value() in ["Guru", "Experimental"] and \
                         "," not in gui.params.param("set", "v").children()[0].name() or \
                         gui.params.param("vis").value() not in ["Guru", "Experimental"] and \
                         "," in gui.params.param("set", "v").children()[0].name() or \
                         is2D and "," not in gui.params.param("set", "v").children()[0].name() or \
                         not is2D and "," in gui.params.param("set", "v").children()[0].name()

        with gui.params.treeChangeBlocker():
            if gui.fringes.SDM:
                gui.params.param("sys", "grid").setLimits(("image", "Cartesian"))
            else:
                gui.params.param("sys", "grid").setLimits(("image", "Cartesian", "polar", "log-polar"))
            gui.params.param("sys", "grid").setValue(gui.fringes.grid)

            gui.params.param("sys", "angle").setValue(gui.fringes.angle)
            if gui.fringes.grid in gui.fringes._grids[2:]:
                gui.params.param("sys", "angle").setDefault(45)
            else:
                gui.params.param("sys", "angle").setDefault(0)

            gui.params.param("sys", "D").setValue(gui.fringes.D)

            gui.params.param("sys", "D", "axis").setValue(("X", "Y")[gui.fringes.axis])
            gui.params.param("sys", "D", "axis").show(gui.fringes.D == 1)

            gui.params.param("vid", "T").setValue(gui.fringes.T)
            gui.params.param("vid", "T").setLimits((1 if gui.visibility == "Guru" else 3, gui.fringes._Tmax))

            gui.params.param("vid", "Y").setLimits((1, min(gui.fringes._Ymax, gui.fringes._Pmax / gui.fringes.X)))
            gui.params.param("vid", "Y").setValue(gui.fringes.Y)

            gui.params.param("vid", "Y").setLimits((1, min(gui.fringes._Xmax, gui.fringes._Pmax / gui.fringes.Y)))
            gui.params.param("vid", "X").setValue(gui.fringes.X)

            gui.params.param("vid", "C").setValue(gui.fringes.C)

            gui.params.param("vid", "alpha").setValue(gui.fringes.alpha)

            gui.params.param("vid", "L").setValue(gui.fringes.L)

            gui.params.param("set", "K").setValue(gui.fringes.K)

            D = gui.fringes.D if is2D else 1
            for d in range(D):
                for k in range(gui.fringes.K):
                    id = str(k).translate(gui.sub)
                    if is2D:  # prepend index for direction
                        id = str(d).translate(gui.sub) + "," + id

                    if change_indices:  # remove and add child params
                        if d == k == 0:  # in first loop: remove child params
                            gui.params.param("set", "N").clearChildren()
                            gui.params.param("set", "l").clearChildren()
                            gui.params.param("set", "v").clearChildren()
                            gui.params.param("set", "f").clearChildren()

                        if gui.fringes.FDM:  # all shifts must be equal; hence create only one generic index "i"
                            if d == k == 0:
                                gui.params.param("set", "N").addChild(
                                    {
                                        "name": "N" + "ᵢ",
                                        "type": "int",
                                        "value": gui.fringes._N[d, k],
                                        "default": gui.fringes.Nmin if gui.fringes.FDM else gui.fringes.defaults["N"][
                                            0, 0],
                                        "limits": (max(gui.fringes.Nmin,
                                                       1 if gui.visibility == "Guru" else 2 if gui.visibility == "Expert" else 3),
                                                   gui.fringes._Nmax),
                                        "tip": gui.fringes.__class__.N.__doc__,
                                    }
                                )
                        else:
                            gui.params.param("set", "N").addChild(
                                {
                                    "name": "N" + id,
                                    "type": "int",
                                    "value": gui.fringes._N[d, k],
                                    "default": gui.fringes.Nmin if gui.fringes.FDM else gui.fringes.defaults["N"][0, 0],
                                    "limits": (max(gui.fringes.Nmin,
                                                   1 if gui.visibility == "Guru" else 2 if gui.visibility == "Expert" else 3),
                                               gui.fringes._Nmax),
                                    "tip": gui.fringes.__class__.N.__doc__,
                                }
                            )

                        gui.params.param("set", "l").addChild(
                            {
                                "title": "\u03BB" + id,
                                "name": "l" + id,
                                "type": "float",
                                "value": gui.fringes._l[d, k],
                                "default": None,  # gui.fringes.L ** (1 / (k + 1)),
                                "limits": (gui.fringes.lmin, None),
                                "decimals": gui.digits,
                                "tip": gui.fringes.__class__.l.__doc__,
                            }
                        )

                        gui.params.param("set", "v").addChild(
                            {
                                "title": "\u03BD" + id,
                                "name": "v" + id,
                                "type": "float",
                                "value": gui.fringes._v[d, k],
                                "default": None,  # gui.fringes.L ** (1 - 1 / (k + 1)),
                                "limits": (0, gui.fringes.vmax),
                                "decimals": gui.digits,
                                "tip": gui.fringes.__class__.v.__doc__,
                            }
                        )

                        gui.params.param("set", "f").addChild(
                            {
                                "name": "f" + id,
                                "type": "float",
                                "value": gui.fringes._f[d, k],
                                "default": gui.fringes._f[
                                    d, k] if gui.fringes.FDM and gui.fringes.static else gui.fringes.D * k + k if gui.fringes.FDM else 1,
                                "limits": (-gui.fringes.vmax, gui.fringes.vmax),
                                "decimals": gui.digits,
                                "readonly": gui.fringes.FDM and gui.fringes.static,
                                "tip": gui.fringes.__class__.f.__doc__,
                            }
                        )
                    else:  # update child params
                        if gui.fringes.FDM:  # all shifts must be equal; hence create only one generic index "i"
                            if d == k == 0:
                                gui.params.param("set", "N", "N" + "ᵢ").setLimits((max(gui.fringes.Nmin,
                                                                                       1 if gui.visibility == "Guru" else 2 if gui.visibility == "Expert" else 3),
                                                                                   gui.fringes._Nmax))
                                gui.params.param("set", "N", "N" + "ᵢ").setValue(gui.fringes._N[d, k])
                                gui.params.param("set", "N", "N" + "ᵢ").setDefault(
                                    gui.fringes.Nmin if gui.fringes.FDM else gui.fringes.defaults["N"][0, 0])
                        else:
                            gui.params.param("set", "N", "N" + id).setLimits((max(gui.fringes.Nmin,
                                                                                  1 if gui.visibility == "Guru" else 2 if gui.visibility == "Expert" else 3),
                                                                              gui.fringes._Nmax))
                            gui.params.param("set", "N", "N" + id).setValue(gui.fringes._N[d, k])
                            gui.params.param("set", "N", "N" + id).setDefault(
                                gui.fringes.Nmin if gui.fringes.FDM else gui.fringes.defaults["N"][0, 0])

                        gui.params.param("set", "l", "l" + id).setLimits((gui.fringes.lmin, None))
                        gui.params.param("set", "l", "l" + id).setValue(gui.fringes._l[d, k])

                        gui.params.param("set", "v", "v" + id).setLimits((0, gui.fringes.vmax))
                        gui.params.param("set", "v", "v" + id).setValue(gui.fringes._v[d, k])

                        gui.params.param("set", "f", "f" + id).setLimits((-gui.fringes.vmax, gui.fringes.vmax))
                        gui.params.param("set", "f", "f" + id).setValue(gui.fringes._f[d, k])
                        gui.params.param("set", "f", "f" + id).setDefault(None if gui.fringes.FDM else 1)

            gui.params.param("set", "reverse").setValue(gui.fringes.reverse)

            gui.params.param("set", "o").setValue(gui.fringes.o / np.pi)

            gui.params.param("set", "lmin").setValue(gui.fringes.lmin)

            gui.params.param("set", "vmax").setValue(gui.fringes.vmax)

            gui.params.param("set", "UMR").setValue(gui.fringes.UMR.min())

            gui.params.param("val", "dtype").setValue(gui.fringes.dtype)

            gui.params.param("val", "Imax").setValue(gui.fringes.Imax)

            gui.params.param("val", "A").setLimits((gui.fringes.Amin, gui.fringes.Amax))
            gui.params.param("val", "A").setValue(gui.fringes.A)
            gui.params.param("val", "A").setDefault(gui.fringes.Imax / 2)

            gui.params.param("val", "B").setLimits((0, gui.fringes.Bmax))
            gui.params.param("val", "B").setValue(gui.fringes.B)
            gui.params.param("val", "B").setDefault(gui.fringes.Imax / 2)

            gui.params.param("val", "V").setValue(gui.fringes.V)

            gui.params.param("val", "gamma").setValue(gui.fringes.gamma)

            if len(gui.params.param("col", "H").children()) != gui.fringes.H or \
                    gui.params.param("col", "H").value() != gui.fringes.H or \
                    gui.params.param("col", "M").value() != gui.fringes.M:  # remove and add cild params

                gui.params.param("col", "H").clearChildren()

                for h in range(gui.fringes.H):
                    id = str(h).translate(gui.sub)
                    gui.params.param("col", "H").addChild(
                        {
                            "name": "h" + id,
                            "type": "color",
                            "value": gui.fringes.h[h],
                            "default": gui.fringes.defaults["h"][0],
                            "limits": (1, gui.fringes._Hmax),
                            "readonly": gui.visibility in ["Beginner", "Expert"]
                        }
                    )
            else:  # update cild params
                for h in range(gui.fringes.H):
                    id = str(h).translate(gui.sub)
                    gui.params.param("col", "H", "h" + id).setValue(gui.fringes.h[h])

            gui.params.param("col", "M").setValue(gui.fringes.M)

            gui.params.param("col", "H").setValue(gui.fringes.H)

            gui.params.param("mux", "TDM").setValue(gui.fringes.TDM)

            gui.params.param("mux", "SDM").setValue(gui.fringes.SDM)
            gui.params.param("mux", "SDM").setReadonly(not gui.SDMOK)

            gui.params.param("mux", "WDM").setValue(gui.fringes.WDM)
            gui.params.param("mux", "WDM").setReadonly(not gui.WDMOK)

            gui.params.param("mux", "FDM").setValue(gui.fringes.FDM)
            gui.params.param("mux", "FDM").setReadonly(not gui.FDMOK)

            gui.params.param("mux", "FDM", "static").setValue(gui.fringes.static)
            gui.params.param("mux", "FDM", "static").show(gui.fringes.FDM)

            gui.params.param("uwr", "PU").setValue(gui.fringes.PU)

            gui.params.param("uwr", "mode").setValue(gui.fringes.mode)

            gui.params.param("uwr", "Vmin").setValue(gui.fringes.Vmin)

            gui.params.param("uwr", "verbose").setValue(gui.fringes.verbose)

            gui.params.param("quali", "eta").setValue(gui.fringes.eta.max())

            gui.params.param("quali", "dark").setValue(gui.fringes.dark)

            gui.params.param("quali", "quant").setValue(gui.fringes.quant)

            gui.params.param("quali", "shot").setValue(gui.fringes.shot)

            gui.params.param("quali", "u").setValue(gui.fringes.u)

            gui.params.param("quali", "DR").setValue(gui.fringes.DRdB.max())

        gui.reset_button.setEnabled(gui.fringes.params != gui.initials)
        gui.encode_button.setStyleSheet("" if gui.encodeOK else "QPushButton{color: red}")
        gui.decode_button.setEnabled(gui.decodeOK)
        gui.decode_key.setEnabled(gui.decodeOK)
    gui.update_parameter_tree = update_parameter_tree

    def set_shifts():
        gui.fringes.N = "auto"
        update_parameter_tree()

    def set_wavelengths():
        gui.fringes.l = "auto"
        update_parameter_tree()

    def set_periods():
        gui.fringes.v = "auto"
        update_parameter_tree()

    def set_frequencies():
        gui.fringes.f = "auto"
        update_parameter_tree()

    gui.params.param("set", "N").sigActivated.connect(set_shifts)
    gui.params.param("set", "l").sigActivated.connect(set_wavelengths)
    gui.params.param("set", "v").sigActivated.connect(set_periods)
    gui.params.param("set", "f").sigActivated.connect(set_frequencies)
    a = gui.params
    b = gui.params.sigTreeStateChanged

    gui.params.sigTreeStateChanged.connect(set_param)



import numpy as np
import os
import functools

from pyqtgraph.Qt import QtWidgets
from PyQt6.QtWidgets import QFileDialog
import pyqtgraph as pg
import cv2
import json
import yaml
import toml
import fringes as frng


config = {
    ".toml": toml.load,
    ".yml": yaml.safe_load,
    ".yaml": yaml.safe_load,
    ".json": json.load,
}

image = {
    ".bmp": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".dip": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),  # i.e. ".bmp"
    # ".jpeg": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".jpg": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".jpe": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".jp2": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    ".png": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".webp": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".pbm": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".pgm": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".ppm": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".pxm": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".pnm": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".sr": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    # ".ras": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    ".tiff": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
    ".tif": functools.partial(cv2.imread, flags=cv2.IMREAD_UNCHANGED),
}

numpy = {
    ".mmap": np.load,
    ".npy": np.load,
    # ".npz": np.load  # todo
}

loader = {**config, **image, **numpy}


def set_logic(gui):
    """Assigns functionality to the buttons and keys in the GUI."""

    def immerse():
        gui.immerse = not gui.immerse

        if gui.immerse:
            gui.win.showFullScreen()
        else:
            gui.win.showMaximized()

    def reset():
        gui.fringes.reset()
        gui.update_parameter_tree()

        clear()

    def load():
        """Load data from given directory."""

        flist = QFileDialog.getOpenFileNames(
            caption="Select file(s)",
            directory=os.path.join(os.path.expanduser("~"), "Videos"),
            # options=QFileDialog.Option.DontUseNativeDialog,
            filter=f"Images {tuple('*' + key for key in image.keys())};;Numpy {tuple('*' + key for key in numpy.keys())};;Config {tuple('*' + key for key in config.keys())}".replace(",", "").replace("'", "")
        )

        if flist[0]:
            with pg.BusyCursor():
                path, base = os.path.split(flist[0][0])
                name, ext = os.path.splitext(base)

                if ext in config.keys():
                    gui.fringes.load(flist[0][0])
                    gui.update_parameter_tree()
                else:  # load data
                    root = name.rstrip("1234567890").rstrip("_")

                    data = loader[ext](flist[0][0])
                    shape = data.shape
                    dtype = data.dtype

                    sequence = np.empty((len(flist[0]),) + shape, dtype)
                    sequence[0] = data
                    for i, f in enumerate(flist[0][1:]):
                        path, base = os.path.split(f)
                        name, ext_i = os.path.splitext(base)
                        root_i = name.rstrip("1234567890").rstrip("_")
                        data_i = loader[ext_i](f)

                        if ext_i == ext and root_i == root and data.shape == data_i.shape and data.dtype == data_i.dtype:
                            sequence[i+1] = data_i
                        else:
                            gui.fringes.logger.error("Files in list dint't match; terminated loading data.")
                            return

                    setattr(gui.con, root, frng.vshape(sequence))
                    gui.fringes.logger.info(f"Loaded data from '{path}'.")

                    view(getattr(gui.con, root))
                    gui.data_table.setData(gui.con.info)
                    QtWidgets.QApplication.processEvents()  # refresh event queue

            gui.decode_button.setEnabled(gui.decodeOK)
            gui.decode_key.setEnabled(gui.decodeOK)
            gui.remap_button.setEnabled(gui.remapOK)
            gui.remap_key.setEnabled(gui.remapOK)
            gui.curvature_button.setEnabled(gui.curvatureOK)
            gui.curvature_key.setEnabled(gui.curvatureOK)
            gui.height_button.setEnabled(gui.heightOK)
            gui.height_key.setEnabled(gui.heightOK)
            gui.clear_button.setEnabled(gui.dataOK)
            gui.clear_key.setEnabled(gui.dataOK)
            gui.set_button.setEnabled(gui.set_dataOK)

    def save():
        """Save all data to current directory."""

        path = QFileDialog.getExistingDirectory(
            caption="Select directory",
            directory=os.path.join(os.path.expanduser("~"), "Videos"),
            # options=QFileDialog.Option.DontUseNativeDialog,
        )

        if os.path.isdir(os.path.abspath(path)):
            with pg.BusyCursor():
                gui.fringes.save(os.path.join(path, "params.yaml"))

                for k, v in gui.con.__dict__.items():
                    if isinstance(v, np.ndarray) and v.size > 0:
                        T, Y, X, C = v.shape = frng.vshape(v).shape
                        color_order = (2, 1, 0, 3) if C == 4 else (2, 1, 0) if C == 3 else 0  # to compensate OpenCV color order
                        color_channels = (1, 3, 4)
                        is_img_shape = v.ndim <= 2 or v.ndim == 3 and v.shape[-1] in color_channels
                        is_vid_shape = v.ndim == 3 or v.ndim == 4 and v.shape[-1] in color_channels
                        is_img_dtype = v.dtype in (bool, np.uint8, np.uint16) or \
                                       v.dtype in (np.float32,) and np.min(v) >= 0 and np.max(v) <= 1  # todo: np.float16, np.float64

                        if is_img_dtype and is_img_shape:  # save as image
                            fname = os.path.join(path, f"{k}.tif")
                            cv2.imwrite(fname, v[..., color_order])
                        elif is_img_dtype and is_vid_shape:  # save as image sequence
                            for t in range(T):
                                fname = os.path.join(path, f"{k}_{str(t + 1).zfill(len(str(T)))}.tif")
                                cv2.imwrite(fname, v[t][..., color_order])
                        else:  # save as numpy array
                            np.save(os.path.join(path, f"{k}.npy"), v)

                gui.fringes.logger.info(f"Saved data to '{path}'.")

    def clear():
        """Clear all data from the gui_util."""

        view(None)

        gui.key = ""

        con_dict = gui.con.__dict__.copy()
        for k, v in con_dict.items():
            if isinstance(v, np.ndarray):
                delattr(gui.con, k)

        gui.data_table.clearContents()
        gui.data_table.setRowCount(0)
        gui.set_button.setEnabled(False)

        gui.decode_button.setEnabled(False)
        gui.decode_key.setEnabled(False)
        gui.remap_button.setEnabled(False)
        gui.remap_key.setEnabled(False)
        gui.curvature_button.setEnabled(False)
        gui.curvature_key.setEnabled(False)
        gui.height_button.setEnabled(False)
        gui.height_key.setEnabled(False)
        gui.clear_button.setEnabled(False)
        gui.clear_key.setEnabled(False)

    def set_data():
        """Set data (to be decoded)."""

        try:
            item = gui.data_table.currentItem().text()  # todo: select first element of selected row

            if hasattr(gui.con, item):
                view(getattr(gui.con, item))
                gui.key = item
                gui.fringes.logger.info(f"Set data to be decoded to '{gui.key}'.")
            else:
                view(None)
                gui.key = None

            gui.decode_button.setEnabled(gui.decodeOK)
            gui.decode_key.setEnabled(gui.decodeOK)
        except Exception:
            pass

    def encode():
        """Encode fringes based on the given parameters."""
        if hasattr(gui.con, "fringes"):
            delattr(gui.con, "fringes")

        gui.data_table.setData(gui.con.info)
        QtWidgets.QApplication.processEvents()  # refresh event queue

        with pg.BusyCursor():
            gui.con.fringes = gui.fringes.encode()

        view(getattr(gui.con, "fringes"))
        gui.data_table.setData(gui.con.info)
        QtWidgets.QApplication.processEvents()  # refresh event queue

        gui.decode_button.setEnabled(gui.decodeOK)
        gui.decode_key.setEnabled(gui.decodeOK)
        gui.clear_button.setEnabled(gui.dataOK)
        gui.clear_key.setEnabled(gui.dataOK)
        gui.set_button.setEnabled(gui.set_dataOK)
    gui.encode = encode

    def decode():
        """Decode encoded or acquired fringes."""
        if hasattr(gui.con, "brightness"):
            delattr(gui.con, "brightness")
        if hasattr(gui.con, "modulation"):
            delattr(gui.con, "modulation")
        if hasattr(gui.con, "registration"):
            delattr(gui.con, "registration")
        if hasattr(gui.con, "phase"):
            delattr(gui.con, "phase")
        if hasattr(gui.con, "residuals"):
            delattr(gui.con, "residuals")
        if hasattr(gui.con, "orders"):
            delattr(gui.con, "orders")

        gui.data_table.setData(gui.con.info)
        QtWidgets.QApplication.processEvents()  # refresh event queue

        if hasattr(gui.con, gui.key):
            I = getattr(gui.con, gui.key)
        # elif hasattr(gui.con, "raw"):
        #     I = gui.con.raw
        elif hasattr(gui.con, "fringes"):
            I = gui.con.fringes

        with pg.BusyCursor():
            dec = gui.fringes.decode(I)

            if dec is not None:
                for k, v in dec._asdict().items():
                    setattr(gui.con, k, v)

            view(getattr(gui.con, "registration"))
        gui.data_table.setData(gui.con.info)
        QtWidgets.QApplication.processEvents()  # refresh event queue

        gui.remap_button.setEnabled(gui.remapOK)
        gui.remap_key.setEnabled(gui.remapOK)
        gui.curvature_button.setEnabled(gui.curvatureOK)
        gui.curvature_key.setEnabled(gui.curvatureOK)
        gui.set_button.setEnabled(gui.set_dataOK)
    gui.decode = decode

    def remap():
        if hasattr(gui.con, "source"):
            delattr(gui.con, "source")

        gui.data_table.setData(gui.con.info)
        QtWidgets.QApplication.processEvents()  # refresh event queue

        with pg.BusyCursor():
            x = gui.con.registration
            B = gui.con.modulation
            Bmin = 0
            scale = 1
            normalize = True

            gui.con.source = gui.fringes.remap(x, B, Bmin, scale, normalize)

            gui.fringes.logger.info("Remapped.")

        view(getattr(gui.con, "source"))
        gui.data_table.setData(gui.con.info)
        QtWidgets.QApplication.processEvents()  # refresh event queue

    def curvature():
        if hasattr(gui.con, "curvature"):
            delattr(gui.con, "curvature")

        gui.data_table.setData(gui.con.info)
        QtWidgets.QApplication.processEvents()  # refresh event queue

        with pg.BusyCursor():
            gui.con.curvature = frng.curvature(gui.con.registration)

            gui.fringes.logger.info("Computed curvature.")

        view(getattr(gui.con, "curvature"))
        gui.data_table.setData(gui.con.info)
        QtWidgets.QApplication.processEvents()  # refresh event queue

        gui.height_button.setEnabled(True)
        gui.height_key.setEnabled(True)

    def height():
        if hasattr(gui.con, "height"):
            delattr(gui.con, "height")

        with pg.BusyCursor():
            gui.con.height = frng.height(gui.con.curvature)

            gui.fringes.logger.info("Computed height.")

        view(getattr(gui.con, "height"))
        gui.data_table.setData(gui.con.info)
        QtWidgets.QApplication.processEvents()  # refresh event queue

        gui.data_table.setData(gui.con.info)
        QtWidgets.QApplication.processEvents()  # refresh event queue

    def view(I=None, Imax=None, autoscale=False, enhance=False, cmap=None):  # todo: color map
        """Display image data in videoshape in the ImageView area of the GUI."""
        try:
            T, Y, X, C = frng.vshape(I).shape
        except:
            T = Y = X = C = None

        if enhance and I is not None and T == 1 and 3 <= C and I.dtype == "uint8":  # improve lightfield-inspection
            try:
                I[..., 1] = 0
                hsv = cv2.cvtColor(I, cv2.COLOR_RGB2HSV)
                hsv[..., 1] = 255  # 2
                I = cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)
            except:
                pass

        if I is None or not isinstance(I, np.ndarray) or not I.size:
            gui.imv.clear()
        elif I.ndim < 2:
            gui.fringes.logger.error("Can't display array with less than 2 dimensions.")  # todo: convert into videoshape
        elif I.ndim > 4:
            gui.fringes.logger.error("Can't display array with more than 4 dimensions.")  # todo: convert into videoshape
        elif I.dtype.kind in "b":  # bool
            gui.imv.setImage(I, autoLevels=False, levels=(0, 1), autoHistogramRange=False)
            # gui.plot.setLimits(xMin=0, xMax=X, yMin=0, yMax=Y)
        elif I.dtype.kind in "ui":  # uint or int
            if autoscale:
                gui.imv.setImage(I)
            else:
                if I.dtype.kind == "u" or np.min(I) >= 0:
                    Imin = 0
                else:
                    Imin = np.iinfo(I.dtype).min

                if Imax is not None:
                    Imax = int(Imax)
                elif I.dtype.itemsize > 1:  # e.g. 16bit data may hold only 10bit or 12bit or 14bit information
                    if np.any(I > 2 ** 10 - 1):  # 10-bit data
                        Imax = 2 ** 10 - 1
                    elif np.any(I > 2 ** 12 - 1):  # 12-bit data
                        Imax = 2 ** 12 - 1
                    elif np.any(I > 2 ** 14 - 1):  # 14-bit data
                        Imax = 2 ** 14 - 1
                    else:
                        Imax = np.iinfo(I.dtype).max
                else:
                    Imax = np.iinfo(I.dtype).max

                gui.imv.setImage(I, autoLevels=False, levels=(Imin, Imax), autoHistogramRange=False)
                # gui.plot.setLimits(xMin=0, xMax=X, yMin=0, yMax=Y)
        elif I.dtype.kind in "f":  # float
            try:
                if np.isnan(np.amax(I)):
                    I = I.astype(np.float32)  # copy
                    I[np.isnan(I)] = 0
                gui.imv.setImage(I.astype(np.float32, copy=False))  # it's faster for float32 than for float64
            except:
                gui.fringes.logger.error("Couldn't display float data.")

            # if cmap:  # todo: define cmaps
            #     try:
            #         from pyqtgraph.graphicsItems.GradientEditorItem import Gradients
            #
            #         # list the available colormaps
            #         print(Gradients.keys())
            #
            #         # pick one to turn into an actual colormap
            #         spectrumColormap = pg.ColorMap(*zip(*Gradients["spectrum"]["ticks"]))
            #
            #         # create colormaps from the builtins
            #         pos, color = zip(*Gradients[cmap]["ticks"])
            #         cmap = pg.ColorMap(pos, color)
            #
            #         gui.imv.setColorMap(cmap)
            #     except:
            #         pass

            # gui.plot.setLimits(xMin=0, xMax=X, yMin=0, yMax=Y)
        else:
            gui.fringes.logger.error("Can't display array with dtype %s." % I.dtype)

        QtWidgets.QApplication.processEvents()  # refresh event queue
    gui.view = view

    def zoomback():
        pass  # todo: zoom back strg + 0

    def selection_changed():
        """Display the data which was selected."""

        try:
            item = gui.data_table.currentItem().text()  # todo: select first element of selected row
        except:
            return

        if hasattr(gui.con, item):
            view(getattr(gui.con, item))
        else:
            view(None)

    # assign functionality to buttons
    gui.immerse_key.activated.connect(immerse)

    gui.reset_button.clicked.connect(reset)
    gui.reset_key.activated.connect(reset)

    gui.load_button.clicked.connect(load)
    gui.load_key.activated.connect(load)
    gui.save_button.clicked.connect(save)
    gui.save_key.activated.connect(save)
    gui.clear_button.clicked.connect(clear)
    gui.clear_key.activated.connect(clear)

    gui.set_button.clicked.connect(set_data)
    gui.set_key.activated.connect(set_data)
    gui.data_table.itemSelectionChanged.connect(selection_changed)

    gui.encode_button.clicked.connect(encode)
    gui.encode_key.activated.connect(encode)
    gui.decode_button.clicked.connect(decode)
    gui.decode_key.activated.connect(decode)

    gui.remap_button.clicked.connect(remap)
    gui.remap_key.activated.connect(remap)
    gui.curvature_button.clicked.connect(curvature)
    gui.curvature_key.activated.connect(curvature)
    gui.height_button.clicked.connect(height)
    gui.height_key.activated.connect(height)

    gui.zoomback_key.activated.connect(zoomback)



import logging as lg

import numpy as np


def set_params(gui):
    """Define the parameter tree."""

    vis = {
        "title": "Visibility",
        "name": "vis",
        "type": "list",
        "value": "Expert",
        "limits": ["Beginner", "Expert", "Guru", "Experimental"],
        "tip": f"The Visibility defines the type of user that should get access to the feature. "
               f"It does not affect the functionality of the features but is used by the GUI to "
               f"decide which features to display based on the current user level. The purpose "
               f"is mainly to ensure that the GUI is not cluttered with information that is not "
               f"intended at the current visibility level. The following criteria have been used "
               f"for the assignment of the recommended visibility:\n"
               f"\u2B9A Beginner: Features that should be visible for all users via the GUI. This "
               f"is the default visibility. The number of features with 'Beginner' visibility "
               f"should be limited to all basic features so the GUI display is well-organized "
               f"and easy to use.\n"
               f"\u2B9A Expert: Features that require a more in-depth knowledge of the system "
               f"functionality. This is the preferred visibility level for all advanced features.\n"
               f"\u2B9A Guru: - Guru: Advanced features that usually only people"
               f"with a sound background in phase shifting can make good use of.\n"
               f"\u2B9A Experimental: New features that have not been tested yet "
               f"and the system might probably crash at some point.",
    }
    log = {
        "title": "Logging",
        "name": "log",
        "type": "list",
        "value": lg.getLevelName(gui.fringes.logger.level),
        "default": lg.getLevelName(gui.fringes.logger.level),
        "limits": ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
        "tip": "Logging level."
    }
    vid = {
        "title": "Video Shape",
        "name": "vid",
        "type": "group",
        "value": gui.fringes.T,
        "limits": [1, gui.fringes._Tmax],
        "tip": gui.fringes.__class__.T.__doc__,
        "children": [
            {
                "title": "Frames",
                "name": "T",
                "type": "int",
                "value": gui.fringes.T,
                "default": gui.fringes.T,
                "limits": (1 if gui.visibility == "Guru" else 3, gui.fringes._Tmax),
                "tip": gui.fringes.__class__.T.__doc__,
            },
            {
                "title": "Height",
                "name": "Y",
                "type": "int",
                "value": gui.fringes.Y,
                "default": gui.fringes.Y,  # defaults["Y"]
                "limits": (
                    1, min(gui.fringes._Ymax, gui.fringes._Pmax / gui.fringes.X) if gui.visibility == "Guru" else 7680
                ),
                "step": 10,
                "suffix": " px",
                "tip": gui.fringes.__class__.Y.__doc__,
            },
            {
                "title": "Width",
                "name": "X",
                "type": "int",
                "value": gui.fringes.X,
                "default": gui.fringes.X,  # defaults["X"]
                "limits": (
                    1, min(gui.fringes._Xmax, gui.fringes._Pmax / gui.fringes.Y) if gui.visibility == "Guru" else 7680
                ),
                "step": 10,
                "suffix": " px",
                "tip": gui.fringes.__class__.X.__doc__,
            },
            {
                "title": "Colors",
                "name": "C",
                "type": "int",
                "value": gui.fringes.C,
                "readonly": True,
                "visible": gui.visibility == "Guru",
                "tip": gui.fringes.__class__.C.__doc__,
            },
            {
                "title": "alpha",
                "name": "alpha",
                "type": "float",
                "value": gui.fringes.alpha,
                "default": gui.fringes.defaults["alpha"],
                "limits": (1, 2),
                "step": 0.1,
                "decimals": gui.digits,
                "visible": gui.visibility == "Guru",
                "tip": gui.fringes.__class__.alpha.__doc__,
            },
            {
                "title": "Length",
                "name": "L",
                "type": "float",
                "value": gui.fringes.L,
                "decimals": gui.digits,
                "suffix": "px",
                "readonly": True,
                "tip": gui.fringes.__class__.L.__doc__,
            },

        ]
    }
    sys = {
            "title": "Coordinates",
            "name": "sys",
            "type": "group",
            "children": [
                {
                    "title": "Grid",
                    "name": "grid",
                    "type": "list",
                    "value": gui.fringes.grid,
                    "default": gui.fringes.defaults["grid"],
                    "limits": gui.fringes._grids,
                    "visible": gui.visibility == "Guru",
                    "tip": gui.fringes.__class__.grid.__doc__,
                },
                {
                    "title": "Angle",
                    "name": "angle",
                    "type": "float",
                    "value": gui.fringes.angle,
                    "default": gui.fringes.defaults["angle"],
                    "limits": (-360, 360),  # todo: +-45
                    "decimals": gui.digits,
                    "suffix": "°",
                    "visible": gui.visibility == "Guru",
                    "tip": gui.fringes.__class__.angle.__doc__,
                },
                {
                    "title": "Directions",
                    "name": "D",
                    "type": "int",
                    "value": gui.fringes.D,
                    "default": gui.fringes.defaults["D"],
                    "limits": (1, gui.fringes._Dmax),
                    "tip": gui.fringes.__class__.D.__doc__,
                    "children": [
                        {
                            "title": "Axis",
                            "name": "axis",
                            "type": "list",
                            "value": ("X", "Y")[gui.fringes.axis],
                            "default": ("X", "Y")[gui.fringes.defaults["axis"]],
                            "limits": ("X", "Y"),
                            "visible": gui.fringes.D == 1,
                            "tip": gui.fringes.__class__.axis.__doc__,
                        },
                    ],
                },
            ]
        }
    set = {
            "title": "Set",
            "name": "set",
            "type": "group",
            "children": [
                {
                    "title": "Sets",
                    "name": "K",
                    "type": "int",
                    "value": gui.fringes.K,
                    "default": gui.fringes.defaults["K"],
                    "limits": (1, gui.fringes._Kmax),
                    # 1 if gui.visibility != "Guru" else gui.fringes._Kmax
                    "tip": gui.fringes.__class__.K.__doc__,
                },
                {
                    "title": "Shifts",
                    "name": "N",
                    "type": "action",
                    "tip": "Reset values to defaults.",
                    "children": [
                        {
                            "name": "N" + str(d).translate(gui.sub) + ", " + str(k).translate(gui.sub),
                            "type": "int",
                            "value": gui.fringes._N[d, k],
                            "default": gui.fringes.Nmin if gui.fringes.FDM else gui.fringes.defaults["N"][0, 0],
                            "limits": (max(gui.fringes.Nmin, 1 if gui.visibility == "Guru" else 2 if gui.visibility == "Expert" else 3), gui.fringes._Nmax),
                            "tip": gui.fringes.__class__.N.__doc__,
                        } for d in range(gui.fringes.D) for k in range(gui.fringes.K)
                    ] if gui.visibility == "Guru" or gui.fringes.N.ndim > 1 else [
                        {
                            "name": "N" + str(k).translate(gui.sub),
                            "type": "int",
                            "value": gui.fringes.N[k],
                            "default": gui.fringes.Nmin if gui.fringes.FDM else gui.fringes.defaults["N"][0, 0],
                            "limits": (max(gui.fringes.Nmin, 1 if gui.visibility == "Guru" else 2 if gui.visibility == "Expert" else 3), gui.fringes._Nmax),
                            "tip": gui.fringes.__class__.N.__doc__,
                        } for k in range(gui.fringes.K)
                    ],
                },
                {
                    "title": "Wavelengths",
                    "name": "l",
                    "type": "action",
                    "tip": "Set optimal wavelengths automatically.",
                    "children": [
                        {
                            "title": "\u03BB" + str(d).translate(gui.sub) + ", " + str(k).translate(gui.sub),
                            "name": "l" + str(d).translate(gui.sub) + ", " + str(k).translate(gui.sub),
                            "type": "float",
                            "value": gui.fringes._l[d, k],
                            "default": None,  # gui.fringes.L ** (1 / (k + 1)),
                            "limits": (gui.fringes.lmin, None),
                            "decimals": gui.digits,
                            "tip": gui.fringes.__class__.l.__doc__,
                        } for d in range(gui.fringes.D) for k in range(gui.fringes.K)
                    ] if gui.visibility == "Guru" or gui.fringes.l.ndim > 1 else [
                        {
                            "title": "\u03BB" + str(k).translate(gui.sub),
                            "name": "l" + str(k).translate(gui.sub),
                            "type": "float",
                            "value": gui.fringes.l[k],
                            "default": None,  # gui.fringes.L ** (1 / (k + 1)),
                            "suffix": " px",
                            "limits": (gui.fringes.lmin, None),
                            "decimals": gui.digits,
                            "tip": gui.fringes.__class__.l.__doc__,
                        } for k in range(gui.fringes.K)
                    ],
                },
                {
                    "title": "Periods",
                    "name": "v",
                    "type": "action",
                    "tip": "Set optimal periods automatically.",
                    "children": [
                        {
                            "title": "\u03BD" + str(d).translate(gui.sub) + ", " + str(k).translate(gui.sub),
                            "name": "v" + str(d).translate(gui.sub) + ", " + str(k).translate(gui.sub),
                            "type": "float",
                            "value": gui.fringes._v[d, k],
                            "default": None,  # gui.fringes.L ** (1 - 1 / (k + 1)),
                            "limits": (0, gui.fringes.vmax),
                            "decimals": gui.digits,
                            "tip": gui.fringes.__class__.v.__doc__,
                        } for d in range(gui.fringes.D) for k in range(gui.fringes.K)] if gui.visibility == "Guru" or gui.fringes.v.ndim > 1 else [
                        {
                            "title": "\u03BD" + str(k).translate(gui.sub),
                            "name": "v" + str(k).translate(gui.sub),
                            "type": "float",
                            "value": gui.fringes.v[k],
                            "default": None,  # gui.fringes.L ** (1 - 1 / (k + 1)),
                            "limits": (0, gui.fringes.vmax),
                            "decimals": gui.digits,
                            "tip": gui.fringes.__class__.v.__doc__,
                        } for k in range(gui.fringes.K)
                    ],
                },
                {
                    "title": "Frequencies",
                    "name": "f",
                    "type": "action",
                    "visible": gui.visibility == "Guru",
                    "tip": "Reset values to defaults.",
                    "children": [
                        {
                            "name": "f" + str(d).translate(gui.sub) + ", " + str(k).translate(gui.sub),
                            "type": "float",
                            "value": gui.fringes._f[d, k],
                            "default": None if gui.fringes.FDM else 1,
                            "limits": (-gui.fringes.vmax, gui.fringes.vmax),
                            "decimals": gui.digits,
                            "readonly": gui.fringes.FDM and gui.fringes.static,
                            "tip": gui.fringes.__class__.f.__doc__,
                        } for d in range(gui.fringes.D) for k in range(gui.fringes.K)
                    ] if gui.visibility == "Guru" or gui.fringes.FDM or gui.fringes.f.ndim > 1 else [
                        {
                            "name": "f" + str(k).translate(gui.sub),
                            "type": "float",
                            "value": gui.fringes.f[k],
                            "default": 1,
                            "limits": (-gui.fringes.vmax, gui.fringes.vmax),
                            "decimals": gui.digits,
                            "tip": gui.fringes.__class__.f.__doc__,
                        } for k in range(gui.fringes.K)
                    ],
                },
                {
                    "title": "Reverse",
                    "name": "reverse",
                    "type": "bool",
                    "value": gui.fringes.reverse,
                    "default": gui.fringes.defaults["reverse"],
                    "tip": gui.fringes.__class__.reverse.__doc__,
                },
                {
                    "title": "Offset",
                    "name": "o",
                    "type": "float",
                    "value": gui.fringes.o / np.pi,
                    "default": gui.fringes.defaults["o"] / np.pi,
                    "limits": (-2, 2),
                    "step": 0.5,
                    "decimals": gui.digits,
                    "suffix": "\U0001D745",  # \U0001D70B
                    "tip": gui.fringes.__class__.o.__doc__,
                },
                {
                    "title": "\u03BB\u2098\u1D62\u2099",
                    "name": "lmin",
                    "type": "int",
                    "value": gui.fringes.lmin,
                    "default": gui.fringes.defaults["lmin"],
                    "limits": (4, gui.fringes.L),
                    "suffix": "px",
                    "tip": gui.fringes.__class__.lmin.__doc__,
                },
                {
                    "title": "\u03BD\u2098\u2090\u2093",
                    "name": "vmax",
                    "type": "float",
                    "value": gui.fringes.vmax,
                    # "default": gui.fringes.defaults["vmax"],
                    "decimals": gui.digits,
                    "readonly": True,
                    "visible": False,  # todo
                    "tip": gui.fringes.__class__.vmax.__doc__,
                },
                {
                    "title": "Range",
                    "name": "UMR",
                    "type": "float",
                    "value": gui.fringes.UMR.min(),
                    "decimals": gui.digits,
                    "suffix": " px",
                    "readonly": True,
                    "tip": gui.fringes.__class__.UMR.__doc__,
                },
            ]
        }
    val = {
            "title": "Values",
            "name": "val",
            "type": "group",
            "children": [
                {
                    "title": "Type",
                    "name": "dtype",
                    "type": "list",
                    "value": gui.fringes.dtype,
                    "default": gui.fringes.defaults["dtype"],
                    "limits": gui.fringes._dtypes,
                    "visible": gui.visibility == "Guru",
                    "tip": gui.fringes.__class__.dtype.__doc__,
                },
                {
                    "title": "I\u2098\u2090\u2093",
                    "name": "Imax",
                    "type": "int",
                    "value": gui.fringes.Imax,
                    "readonly": True,
                    "visible": gui.visibility == "Guru",
                    "tip": gui.fringes.__class__.Imax.__doc__,
                },
                {
                    "title": "Bias",
                    "name": "A",
                    "type": "float",
                    "value": gui.fringes.A,
                    "default": gui.fringes.defaults["A"],
                    "limits": (gui.fringes.B, min(gui.fringes.A, gui.fringes.Imax - gui.fringes.A)),
                    "decimals": gui.digits,
                    "tip": gui.fringes.__class__.A.__doc__,
                },
                {
                    "title": "Amplitude",
                    "name": "B",
                    "type": "float",
                    "value": gui.fringes.B,
                    "default": gui.fringes.defaults["B"],
                    "limits": (0, gui.fringes.Imax - gui.fringes.B),
                    "decimals": gui.digits,
                    "tip": gui.fringes.__class__.B.__doc__,
                },
                {
                    "title": "Visibility",
                    "name": "V",
                    "type": "float",
                    "value": gui.fringes.V,
                    "default": gui.fringes.defaults["V"],
                    "limits": (0, 1),
                    "step": 0.1,
                    "decimals": gui.digits,
                    "tip": gui.fringes.__class__.V.__doc__,
                },
                {
                    "title": "Gamma",
                    "name": "gamma",
                    "type": "float",
                    "value": gui.fringes.gamma,
                    "default": gui.fringes.defaults["gamma"],
                    "limits": (0, gui.fringes.__class__._gammamax),
                    "step": 0.1,
                    "decimals": gui.digits,
                    "tip": gui.fringes.__class__.gamma.__doc__,
                },
            ],
        }
    col = {
        "title": "Color",
        "name": "col",
        "type": "group",
        "visible": gui.visibility == "Guru",
        "children": [
            {
                "title": "Averaging",
                "name": "M",
                "type": "float",
                "value": gui.fringes.M,
                "default": gui.fringes.defaults["M"],
                "limits": (1 / 255, gui.fringes._Mmax),
                "tip": gui.fringes.__class__.M.__doc__,
            },
            {
                "title": "Hues",
                "name": "H",
                "type": "int",
                "value": gui.fringes.H,
                "default": gui.fringes.defaults["H"],
                "limits": (1, gui.fringes._Hmax),
                "tip": gui.fringes.__class__.H.__doc__,
                "children": [
                    {
                        "name": "h" + str(h).translate(gui.sub),
                        "type": "color",
                        "value": gui.fringes.h[h],
                        "default": gui.fringes.defaults["h"][0],
                    } for h in range(gui.fringes.H)
                ],
            },
        ],
    }
    mux = {
            "title": "Multiplexing",
            "name": "mux",
            "type": "group",
        "visible": gui.visibility == "Guru",
            "tip": "Multiplexing method.",
            "children": [
                {
                    "title": "TDM",
                    "name": "TDM",
                    "type": "bool",
                    "value": gui.fringes.TDM,
                    "default": gui.fringes.defaults["TDM"],
                    "readonly": True,
                    "tip": gui.fringes.__class__.TDM.__doc__,
                },
                {
                    "title": "SDM",
                    "name": "SDM",
                    "type": "bool",
                    "value": gui.fringes.SDM,
                    "default": gui.fringes.defaults["SDM"],
                    "tip": gui.fringes.__class__.SDM.__doc__,
                },
                {
                    "title": "WDM",
                    "name": "WDM",
                    "type": "bool",
                    "value": gui.fringes.WDM,
                    "default": gui.fringes.defaults["WDM"],
                    "tip": gui.fringes.__class__.WDM.__doc__,
                },
                {
                    "title": "FDM",
                    "name": "FDM",
                    "type": "bool",
                    "value": gui.fringes.FDM,
                    "default": gui.fringes.defaults["FDM"],
                    "tip": gui.fringes.__class__.FDM.__doc__,
                    "children": [
                        {
                            "title": "Static",
                            "name": "static",
                            "type": "bool",
                            "value": gui.fringes.static,
                            "default": gui.fringes.defaults["static"],
                            "visible": gui.fringes.FDM,
                            "tip": gui.fringes.__class__.static.__doc__,
                        },
                    ]
                },
            ],
        }
    uwr = {
            "title": "Unwrapping",
            "name": "uwr",
            "type": "group",
            "children": [
                {
                    "title": "Method",
                    "name": "PU",
                    "type": "str",
                    "value": gui.fringes.PU,
                    "readonly": True,
                    "tip": gui.fringes.__class__.PU.__doc__,
                },
                {
                    "title": "Mode",
                    "name": "mode",
                    "type": "list",
                    "value": gui.fringes.mode,
                    "default": gui.fringes.defaults["mode"],
                    "limits": gui.fringes._modes,
                    "visible": gui.visibility == "Guru",
                    "tip": gui.fringes.__class__.mode.__doc__,
                },
                {
                    "title": "V\u2098\u1D62\u2099",  # todo: Vmin
                    "name": "Vmin",
                    "type": "float",
                    "value": gui.fringes.Vmin,
                    "default": gui.fringes.defaults["Vmin"],
                    "limits": (0, 1),
                    "step": 0.1,
                    "decimals": gui.digits,
                    "visible": gui.visibility == "Guru",
                    "tip": gui.fringes.__class__.Vmin.__doc__
                },
                {
                    "title": "Verbose",
                    "name": "verbose",
                    "type": "bool",
                    "value": gui.fringes.verbose,
                    "default": gui.fringes.defaults["verbose"],
                    "tip": gui.fringes.__class__.verbose.__doc__,
                },
            ]
        }
    quali = {
            "title": "Quality",
            "name": "quali",
            "type": "group",
            "visible": gui.visibility in ["Expert", "Guru"],
            "children": [
                {
                    "title": "Efficiency",  # \u03B7
                    "name": "eta",
                    "type": "float",
                    "value": gui.fringes.eta.max(),
                    "decimals": gui.digits,
                    "readonly": True,
                    "tip": gui.fringes.__class__.eta.__doc__,
                },
                {
                    "title": "Dark Noise",
                    "name": "dark",
                    "type": "float",
                    "value": gui.fringes.dark,
                    "defaults": gui.fringes.defaults["dark"],
                    "limits": (0, np.sqrt(gui.fringes.Imax)),
                    "step": 0.1,
                    "decimals": gui.digits,
                    "visible": gui.visibility == "Guru",
                    "tip": gui.fringes.__class__.dark.__doc__
                },
                {
                    "title": "Quantization Noise",
                    "name": "quant",
                    "type": "float",
                    "value": gui.fringes.quant,
                    "readonly": True,
                    "decimals": gui.digits,
                    "tip": gui.fringes.__class__.quant.__doc__
                },
                {
                    "title": "Shot Noise",
                    "name": "shot",
                    "type": "float",
                    "value": gui.fringes.shot,
                    "defaults": 0,
                    "decimals": gui.digits,
                    "readonly": True,
                    "visible": gui.visibility == "Guru",
                    "tip": gui.fringes.__class__.shot.__doc__
                },
                {
                    "title": "Uncertainty\u2098\u1D62\u2099",
                    "name": "u",
                    "type": "float",
                    "value": gui.fringes.u,
                    "decimals": gui.digits,
                    "suffix": "px",
                    "readonly": True,
                    "tip": gui.fringes.__class__.u.__doc__
                },
                {
                    "title": "Dynamic Range",
                    "name": "DR",
                    "type": "float",
                    "value": gui.fringes.DRdB.max(),
                    "decimals": gui.digits,
                    "suffix": "dB",
                    "readonly": True,
                    "tip": gui.fringes.__class__.DRdB.__doc__,
                },
            ]
        }

    gui.params = [vis, log, vid, sys, set, val, col, mux, uwr, quali]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fringes_gui']

package_data = \
{'': ['*']}

install_requires = \
['fringes>=0.1.5,<0.2.0',
 'opencv-contrib-python>=4.7.0,<5.0.0',
 'pyqt6>=6.4.2,<7.0.0',
 'pyqtgraph>=0.13.2,<0.14.0',
 'pyyaml>=6.0,<7.0',
 'scikit-image>=0.19.3,<0.20.0',
 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'fringes-gui',
    'version': '0.1.0',
    'description': "Simple graphical user interface for the 'fringes' package.",
    'long_description': "# Fringes-GUI\nAuthor: Christian Kludt\n\n## Description\nGraphical user interface for the [fringes](https://pypi.org/project/fringes/) package.\n\n## Installation\nYou can install `fringes-gui` directly from [PyPi](https://pypi.org/project/fringes-gui) via `pip`:\n\n```\npip install fringes-gui\n```\n\n## Usage\nYou import the `fringes-gui` package and call the function `run()`.\n\n```python\nimport fringes_gui as fgui\nfgui.run()\n```\n\nNow the graphical user interface should appear:\n![Screenshot](docs/GUI.png)\\\nScreenshot of the GUI.\n\n### Parameter Tree\nIn the top left corner the attribute widget is located.\nIt contains the parameter tree with which all the properties of the `Fringes` class\nfrom the [fringes](https://pypi.org/project/fringes/) package can be controlled.\nCheck out its website for more details.\nHowever, if you select a parameter and hover over it, a tool tip will appear\ncontaining the docstring of the respective property of the `Fringes` class.\n\nThe Visibility defines the type of user that should get access to the feature.\nIt does not affect the functionality of the features but is used by the GUI to\ndecide which features to display based on the current user level. The purpose\nis mainly to ensure that the GUI is not cluttered with information that is not\nintended at the current visibility level. The following criteria have been used\nfor the assignment of the recommended visibility:\n- Beginner: Features that should be visible for all users via the GUI. This\nis the default visibility. The number of features with 'Beginner' visibility\nshould be limited to all basic features so the GUI display is well-organized\nand easy to use.\n- Expert: Features that require a more in-depth knowledge of the system\nfunctionality. This is the preferred visibility level for all advanced features.\n- Guru: Advanced features that usually only people with a sound background in phase shifting can make good use of.\n- Experimental: New features that have not been tested yet\nand the system might probably crash at some point.\n\n### Function Buttons\nIn the bottem left corner you will find buttons for the associated methods of the `Fringes` class.\nAlternatively, you can use the keyboard shortcuts which are displayed when you hover over the buttons.\nThe buttons are only active if the necessary data has been enoded, decoded or loaded.\n\n### Viewer\nIn the center resides the viewer.\nIf float data is to be displayed, nan is replaced by zeros.\n\n### Data\nIn the top right corner the data widget is located.\nIt lists the data which has been encoded, decoded or was loaded.\n\nIn order to keep the [Parameter Tree](#parameter-tree) consistent with the data,\nonce a parameter has changed, certain data will be removed\nand also certain [buttons](#function-buttons) will be deactivated.\nAs a consequence, if you load data - e.g. the acquired (deflected) fringe pattern sequence - \nthe first element of its videoshape has to match the parameter `Frames` in order to be able to decode it.\n\nTo display any datum listed in the table in the [Viewer](#viewer), simly select the name of it in the table.\n\nKlick the `Load` button to choose a data or parameter set to load.\nWith the `Save` button, all data including the parameter set are saved to the selected directory.\nUse the `Clear all` button to delete all data.\n\nPlease note: By default, the datum `fringes` is decoded.\nIf you want to decode a datum with a different name (e.g. one that you just loaded),\nselect its name in the table and klick `Set data (to be decoded)`.\n\n### Log\nHere the logging of the `Fringes` class is displayed.\nThe logging level can be set in the [Parameter Tree](#parameter-tree).\n\n## License\nCreative Commons Attribution-NonCommercial-ShareAlike 4.0 International Public License\n",
    'author': 'Christian Kludt',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



from gui import FringesGUI


def main():
    gui = FringesGUI()  # todo: encapsulate this with a with statement to ensure params get saved upon shutdown and/or crash
    gui.show()


if __name__ == "__main__":
    main()
